import asyncio

import pytest


@pytest.fixture(scope="session")
def event_loop():
    """Redefine event_loop fixture with session scope"""

    loop = asyncio.get_event_loop_policy().new_event_loop()
    yield loop
    loop.close()



# pylint: disable=invalid-name
import collections

INFINITY = float("inf")


def find_matching(G):
    """Find the most pairs in a bipartite graph.

    The problem is better known as maximum cardinality matching. The bipartite
    graph `G` is described as a Mapping, where the keys are the vertices of one
    set (U) and the values are a Sequence of vertices of the other set (V),
    which describe the edges (E) of the graph. For example the graph with
    U=(U0, U1), V=(V0, V1) and E=((U0, V0), (U0, V1), (U1, V0)) can be written
    as:

        G = {
            'U0': ['V0', 'V1'],
            'U1': ['V0'],
        }

    The return value is a maximum matching M, described as a Mapping from all
    matched vertices in U to their matched vertex in V. For the example above,
    the return value would be:

        M = {
            'U0': 'V1',
            'U1': 'V0',
        }
    """
    # This is an implementation of the Hopcroft-Karp algorithm.

    # M is the current matching. It starts as a partial matching and is updated
    # until it is a maximum matching.
    M = {}
    M_reverse = {}
    layer = {}

    # This breadth-first search finds the shortest augmenting paths. An
    # augmenting path is a special path with the following rules:
    # - the path starts at a free vertex in U
    # - the path can only traverse unmatched edges from U to V
    # - the path can only traverse matched edges from V to U
    # - the path ends at a free vertex in V
    # The search saves the layer of each vertex in U, at which it was
    # encountered in the search, to guide the following depth-first search.
    def breadth_first_search():
        queue = collections.deque()

        # find free vertices in U to use as starting points
        for u in G:
            if u in M:
                layer[u] = INFINITY
            else:
                layer[u] = 0
                queue.append(u)
        layer[None] = INFINITY

        def is_shortest_path(u):
            return layer[u] < layer[None]

        while queue:
            u = queue.popleft()
            if is_shortest_path(u):
                for v in G[u]:
                    # Go from v to u over a matched edge. next_u is None, if v
                    # is free.
                    next_u = M_reverse.get(v)
                    if layer[next_u] is INFINITY:  # if not visited, yet
                        layer[next_u] = layer[u] + 1
                        queue.append(next_u)
        return layer[None] is not INFINITY  # did we reach a free v?

    # This depth-first search is guided by the layers found in the
    # breadth-first search to find the shortest augmenting paths and update the
    # matching M along the way. All previously matched edges in the path are
    # replaced by the unmatched edges in the path. Since the augmenting paths
    # start and end at a free vertex, every found path increases the number of
    # pairs by one.
    def depth_first_search(u):
        for v in G[u]:
            # Go from v to u over a matched edge. next_u is None, if v is free.
            next_u = M_reverse.get(v)
            if layer[next_u] == layer[u] + 1:
                if next_u is None or depth_first_search(next_u):
                    M[u], M_reverse[v] = v, u
                    return True

        # No path found for this u. Mark it, to not try again.
        layer[u] = INFINITY
        return False

    while breadth_first_search():
        # At least one augmenting path was found. Start a depth-first search at
        # every free vertex in U.
        for u in G:
            if u not in M:
                depth_first_search(u)

    return M


def _main():
    G = {
        "U0": ["V0", "V1"],
        "U1": ["V0", "V4"],
        "U2": ["V2", "V3"],
        "U3": ["V0", "V4"],
        "U4": ["V1", "V3"],
    }

    M = find_matching(G)
    for k, v in M.items():
        print(f"{k} -> {v}")


if __name__ == "__main__":
    _main()



#!/usr/bin/env python3

import asyncio
import contextlib
import os
import pathlib
import sys

import not_my_board._jsonrpc as jsonrpc


async def reserve(name, with_name=None):
    spec_file = _find_spec_file(name)
    spec_name = spec_file.stem if not with_name else with_name

    async with agent_proxy() as proxy:
        await proxy.reserve(spec_name, spec_file.as_posix())


async def return_reservation(name):
    async with agent_proxy() as proxy:
        await proxy.return_reservation(name)


async def attach(name, keep_others=False):
    async with agent_proxy() as proxy:
        reserved_names = {e["place"] for e in await proxy.list()}
        if name in reserved_names:
            await proxy.attach(name)

            others = reserved_names - {name}
            if not keep_others and others:
                for other in others:
                    await proxy.return_reservation(name=other, force=True)
        else:
            spec_file = _find_spec_file(name)
            spec_name = spec_file.stem
            await proxy.reserve(spec_name, spec_file.as_posix())
            await proxy.attach(spec_name)

            if not keep_others and reserved_names:
                for other in reserved_names:
                    await proxy.return_reservation(name=other, force=True)


async def detach(name, keep=False):
    async with agent_proxy() as proxy:
        await proxy.detach(name)
        if not keep:
            await proxy.return_reservation(name)


async def list_():
    async with agent_proxy() as proxy:
        return await proxy.list()


async def status():
    async with agent_proxy() as proxy:
        return await proxy.status()


async def uevent(devpath):
    # devpath has a leading "/", so joining with the / operator doesn't
    # work
    sysfs_path = pathlib.Path("/sys" + devpath)
    busnum = (sysfs_path / "busnum").read_text().rstrip()
    devpath = (sysfs_path / "devpath").read_text().rstrip()

    busid = f"{busnum}-{devpath}"

    pipe = pathlib.Path("/run/usbip-refresh-" + busid)
    if pipe.exists():
        print(f"Binding to usbip-host: {busid}", file=sys.stderr)
        match_busid_path = pathlib.Path("/sys/bus/usb/drivers/usbip-host/match_busid")
        if not match_busid_path.exists():
            await _exec("modprobe", "usbip-host")
        match_busid_path.write_text(f"add {busid}")
        bind_path = pathlib.Path("/sys/bus/usb/drivers/usbip-host/bind")
        bind_path.write_text(busid)
        with pipe.open("r+b", buffering=0) as f:
            f.write(b".")
    else:
        print(f"Loading default driver: {busid}", file=sys.stderr)
        probe_path = pathlib.Path("/sys/bus/usb/drivers_probe")
        try:
            probe_path.write_text(busid)
        except OSError:
            # fails for USB Hubs
            pass


async def _exec(*args, **kwargs):
    proc = await asyncio.create_subprocess_exec(*args, **kwargs)
    await proc.communicate()
    if proc.returncode:
        raise RuntimeError(f"{args!r} exited with {proc.returncode}")


def _find_spec_file(name):
    if "/" in name:
        spec_file = pathlib.Path(name)
    else:
        path = pathlib.Path()
        home = pathlib.Path.home()

        while path != home:
            spec_file = path / ".not-my-board" / "specs" / f"{name}.toml"
            if spec_file.is_file():
                break

            if path != path.parent:
                path = path.parent
            else:
                # we're at '/', stop loop
                path = home
        else:
            config_home = pathlib.Path(
                os.environ.get("XDG_CONFIG_HOME", home / ".config")
            )
            spec_file = config_home / "not-my-board" / "specs" / f"{name}.toml"
            if not spec_file.is_file():
                raise ValueError(f"No spec file exists for name {name}")

    return spec_file


@contextlib.asynccontextmanager
async def agent_proxy():
    runtime_dir = pathlib.Path(os.environ["XDG_RUNTIME_DIR"])
    reader, writer = await asyncio.open_unix_connection(
        runtime_dir / "not-my-board.sock"
    )

    async def send(data):
        writer.write(data + b"\n")
        await writer.drain()

    async with jsonrpc.Proxy(send, reader) as proxy:
        yield proxy



import asyncio
import collections
import contextlib
import pathlib
import sys

import pytest

import not_my_board._util as util


class _VM:
    _name = ""

    async def __aenter__(self):
        async with contextlib.AsyncExitStack() as stack:
            await stack.enter_async_context(
                sh_task(f"./scripts/vmctl run {self._name}", f"vm {self._name}")
            )

            self._stack = stack.pop_all()
            await self._stack.__aenter__()
            return self

    async def __aexit__(self, exc_type, exc, tb):
        await self._stack.__aexit__(exc_type, exc, tb)

    async def configure(self):
        await sh(
            f"./scripts/vmctl configure {self._name}", prefix=f"configure {self._name}"
        )

    def ssh_task(self, cmd, *args, **kwargs):
        return sh_task(
            f"./scripts/vmctl ssh {self._name} while-stdin " + cmd,
            *args,
            terminate=False,
            **kwargs,
        )

    def ssh_task_root(self, cmd, *args, **kwargs):
        return sh_task(
            f"./scripts/vmctl ssh {self._name} doas while-stdin " + cmd,
            *args,
            terminate=False,
            **kwargs,
        )

    async def ssh(self, cmd, *args, **kwargs):
        return await sh(f"./scripts/vmctl ssh {self._name} " + cmd, *args, **kwargs)

    async def ssh_poll(self, cmd, timeout=None):
        return await sh_poll(f"./scripts/vmctl ssh {self._name} " + cmd, timeout)


class ServerVM(_VM):
    _name = "server"
    ip = "192.168.200.1"


class ExporterVM(_VM):
    _name = "exporter"
    ip = "192.168.200.2"

    async def usb_attach(self):
        await sh("./scripts/vmctl usb attach")

    async def usb_detach(self):
        await sh("./scripts/vmctl usb detach")


class ClientVM(_VM):
    _name = "client"
    ip = "192.168.200.3"


VMs = collections.namedtuple("VMs", ["server", "exporter", "client"])


@pytest.fixture(scope="session")
async def vms():
    async with ServerVM() as server:
        while True:
            try:
                async with util.connect("127.0.0.1", 5001):
                    pass
                async with util.connect("127.0.0.1", 5002):
                    pass
            except ConnectionRefusedError:
                await asyncio.sleep(0.1)
                continue
            break
        async with ExporterVM() as exporter:
            async with ClientVM() as client:
                await util.run_concurrently(
                    server.configure(),
                    exporter.configure(),
                    client.configure(),
                )
                await exporter.usb_attach()
                yield VMs(server, exporter, client)


async def test_raw_usb_forwarding(vms):
    async with vms.exporter.ssh_task_root(
        "python3 -m not_my_board._usbip export 2-1", "usbip export"
    ):
        # wait for listening socket
        await vms.exporter.ssh_poll("nc -z 127.0.0.1 3240")

        async with vms.client.ssh_task(
            f"python3 -m not_my_board._usbip import {vms.exporter.ip} 2-1 0",
            "usbip import",
        ):
            # wait for USB device to appear
            await vms.client.ssh_poll("test -e /dev/usbdisk")

            await vms.client.ssh("doas mount /media/usb")
            try:
                result = await vms.client.ssh("cat /media/usb/hello")
                assert result.stdout == "Hello, World!"
            finally:
                await vms.client.ssh("doas umount /media/usb")

    await vms.client.ssh("! test -e /sys/bus/usb/devices/2-1")


async def test_usb_forwarding(vms):
    async with vms.server.ssh_task("not-my-board serve", "serve"):
        # wait for listening socket
        await vms.server.ssh_poll("nc -z 127.0.0.1 2092")

        async with vms.exporter.ssh_task_root(
            f"not-my-board export http://{vms.server.ip}:2092 ./src/tests/qemu-usb-place.toml",
            "export",
        ):
            await vms.client.ssh("""'rm -f "$XDG_RUNTIME_DIR/not-my-board.sock"'""")
            async with vms.client.ssh_task(
                f"not-my-board agent http://{vms.server.ip}:2092", "agent"
            ):
                # wait until exported place is registered
                await vms.client.ssh_poll(
                    "wget -q -O - http://192.168.200.1:2092/api/v1/places | grep -q qemu-usb"
                )
                # wait until agent is ready
                await vms.client.ssh_poll(
                    """'test -e "$XDG_RUNTIME_DIR/not-my-board.sock"'"""
                )

                await vms.client.ssh("not-my-board attach ./src/tests/qemu-usb.toml")
                # TODO attach still returns before the device is available.
                # would be nice if it blocks until the device is ready.
                await vms.client.ssh_poll("test -e /sys/bus/usb/devices/2-1")
                try:
                    await vms.exporter.usb_detach()
                    await vms.client.ssh_poll("! test -e /sys/bus/usb/devices/2-1")
                finally:
                    await vms.exporter.usb_attach()

                await vms.client.ssh_poll("test -e /sys/bus/usb/devices/2-1")
                await vms.client.ssh("not-my-board detach qemu-usb")
                await vms.client.ssh("! test -e /sys/bus/usb/devices/2-1")


ShResult = collections.namedtuple("ShResult", ["stdout", "stderr", "returncode"])


async def sh(cmd, check=True, strip=True, prefix=None):
    proc = await asyncio.create_subprocess_shell(
        cmd, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
    )

    stdout, _ = await util.run_concurrently(
        proc.stdout.read(), _log_output(proc.stderr, cmd, prefix)
    )
    await proc.wait()
    if check and proc.returncode:
        raise RuntimeError(f"{cmd!r} exited with {proc.returncode}")

    stdout = stdout.decode("utf-8")
    if strip:
        stdout = stdout.rstrip()

    return ShResult(stdout, None, proc.returncode)


@contextlib.asynccontextmanager
async def sh_task(cmd, prefix=None, terminate=True):
    # need to exec, otherwise only the shell process is killed with
    # proc.terminate()
    proc = await asyncio.create_subprocess_shell(
        f"exec {cmd}",
        stdin=asyncio.subprocess.PIPE,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.STDOUT,
    )

    logging_task = None
    try:
        logging_task = asyncio.create_task(_log_output(proc.stdout, cmd, prefix))
        yield
    finally:
        proc.stdin.close()
        await proc.stdin.wait_closed()

        if terminate:
            with contextlib.suppress(ProcessLookupError):
                proc.terminate()

        await proc.wait()
        if logging_task:
            await logging_task


async def sh_poll(cmd, timeout=None):
    if pathlib.Path("/dev/kvm").exists():
        if timeout is None:
            timeout = 7
        interval = 0.1
    else:
        if timeout is None:
            timeout = 60
        interval = 1

    async def poll_loop():
        while True:
            result = await sh(cmd, check=False)
            if result.returncode == 0:
                break
            await asyncio.sleep(interval)

    await asyncio.wait_for(poll_loop(), timeout)


async def _log_output(stream, cmd, prefix):
    if prefix is None:
        prefix = f"[{cmd}] ".encode()
    else:
        prefix = f"[{prefix}] ".encode()

    async for line in stream:
        sys.stderr.buffer.write(prefix + line)
        sys.stderr.buffer.flush()



#!/usr/bin/env python3

import asyncio
import functools
import json
import logging
import textwrap
import traceback

import not_my_board._util as util

logger = logging.getLogger(__name__)


CODE_INTERNAL_ERROR = -32603
CODE_PARSE_ERROR = -32700
CODE_INVALID_REQUEST = -32600
CODE_METHOD_NOT_FOUND = -32601


class RemoteError(Exception):
    def __init__(self, code, message, data):
        if "traceback" in data:
            details = textwrap.indent(data["traceback"], " " * 4).rstrip()
            super().__init__(f"{message}\n{details}")
        else:
            super().__init__(message)
        self.code = code
        self.data = data


class Server:
    def __init__(self, send, receive_iter, api_obj):
        super().__init__()
        self._send = send
        self._receive_iter = receive_iter
        self._api_obj = api_obj
        self._tasks = set()
        self._tasks_by_id = {}

    async def serve_forever(self):
        try:
            async for raw_data in self._receive_iter:
                task = asyncio.create_task(self._receive(raw_data))
                self._tasks.add(task)
                task.add_done_callback(self._tasks.discard)
        finally:
            await util.cancel_tasks(self._tasks.copy())

    async def _receive_task(self, raw_data):
        try:
            await self._receive(raw_data)
        except Exception:
            traceback.print_exc()

    async def _receive(self, raw_data):
        id_ = None
        next_error = (CODE_PARSE_ERROR, "Parse Error")
        try:
            id_, data = Request.parse_id(raw_data)

            next_error = (CODE_INVALID_REQUEST, "Invalid Request")
            request = Request.from_data(data)

            next_error = (CODE_METHOD_NOT_FOUND, "Method not found")
            assert not request.method.startswith("_")
            if request.method == "rpc.cancel":
                method = self._cancel
            else:
                method = getattr(self._api_obj, request.method)
                logger.info("Method call: %s", request.method)

            next_error = CODE_INTERNAL_ERROR, None
            if id_ is not None:
                self._tasks_by_id[id_] = asyncio.current_task()
                result = await method(*request.args, **request.kwargs)
                response = Response(result, id_)
                await self._send(bytes(response))
            else:
                await method(*request.args, **request.kwargs)
        except Exception as e:
            if id_ is not None:
                code, message = next_error
                if message is None:
                    message = str(e)
                response = ErrorResponse.with_traceback(code, message, id_)
                await self._send(bytes(response))
            else:
                raise
        finally:
            if id_ in self._tasks_by_id:
                del self._tasks_by_id[id_]

    async def _cancel(self, id_):
        if id_ in self._tasks_by_id:
            await util.cancel_tasks([self._tasks_by_id[id_]])


class Proxy:
    def __init__(self, send, receive_iter):
        self._send = send
        self._receive_iter = receive_iter
        self._next_id = 1
        self._pending = {}
        self._is_receiving = True

    async def __aenter__(self):
        self._task = asyncio.create_task(self.io_loop())
        return self

    async def __aexit__(self, exc_type, exc, tb):
        self._task.cancel()
        try:
            await self._task
        except asyncio.CancelledError:
            pass

    async def io_loop(self):
        try:
            async for raw_data in self._receive_iter:
                await self._receive(raw_data)
        finally:
            self._is_receiving = False
            for future in self._pending.values():
                if not future.done():
                    future.set_exception(RuntimeError("Connection closed"))

    async def _receive(self, raw_data):
        id_ = None
        try:
            id_, data = Response.parse_id(raw_data)
            response = Response.from_data(data)

            future = self._pending.get(id_)
            if future and not future.done():
                future.set_result(response.result)
        except Exception as e:
            future = self._pending.get(id_)
            if future and not future.done():
                future.set_exception(e)
            else:
                traceback.print_exc()

    def __getattr__(self, method_name):
        if method_name.startswith("_"):
            raise AttributeError(f"invalid attribute '{method_name}'")
        return functools.partial(self._call, method_name)

    async def _call(self, method_name, *args, **kwargs):
        if not self._is_receiving:
            raise RuntimeError("IO loop is not running, can't receive responses")

        if kwargs.pop("_notification", False):
            id_ = None
        else:
            id_ = self._next_id
            self._next_id += 1

        assert not args or not kwargs, "use either args or kwargs"

        request = Request(method_name, args or kwargs, id_)
        logger.info("Calling: %s", request.method)

        if id_ is not None:
            future = asyncio.get_running_loop().create_future()
            self._pending[id_] = future
            try:
                await self._send(bytes(request))
                return await self._pending[id_]
            except asyncio.CancelledError:
                await self._cancel(id_, request.method)
                raise
            finally:
                del self._pending[id_]
        else:
            await self._send(bytes(request))

    async def _cancel(self, to_cancel_id, to_cancel_name):
        id_ = self._next_id
        self._next_id += 1

        request = Request("rpc.cancel", [to_cancel_id], id_)
        logger.info("Canceling: %s", to_cancel_name)

        future = asyncio.get_running_loop().create_future()
        self._pending[id_] = future
        try:
            await self._send(bytes(request))
            await self._pending[id_]
        # don't request to cancel the cancellation if this task is canceled
        finally:
            del self._pending[id_]


class Message:
    _is_id_required = False
    _body = {}

    @classmethod
    def parse_id(cls, raw_data):
        data = json.loads(raw_data)
        if cls._is_id_required or "id" in data:
            assert isinstance(data["id"], (str, int)), '"id" must be a string or number'
        return data.get("id"), data

    def __bytes__(self):
        return json.dumps(
            {
                "jsonrpc": "2.0",
                **self._body,
            }
        ).encode()


class Request(Message):
    def __init__(self, method, params, id_=None):
        self.method = method
        self.id = id_
        self._body = {
            "method": method,
            "params": params,
        }

        if id_:
            self._body["id"] = id_

        if isinstance(params, list):
            self.args = params
            self.kwargs = {}
        else:
            self.args = []
            self.kwargs = params

    @classmethod
    def from_data(cls, data):
        method = data["method"]
        assert isinstance(method, str), '"method" must be a string'
        params = data.get("params", [])
        assert isinstance(params, (list, dict)), '"params" must be a structured value'
        return cls(method, params, data.get("id"))


class Response(Message):
    _is_id_required = True

    def __init__(self, result, id_):
        self._body = {"result": result, "id": id_}
        self.result = result
        self.id = id_

    @classmethod
    def from_data(cls, data):
        if "error" in data:
            error = data["error"]
            raise RemoteError(error["code"], error["message"], error["data"])
        return cls(data["result"], data["id"])


class ErrorResponse(Response):
    # pylint: disable=super-init-not-called
    def __init__(self, code, message, id_, data=None):
        self._body = {
            "error": {
                "code": code,
                "message": message,
            },
            "id": id_,
        }
        if data is not None:
            self._body["error"]["data"] = data

    @classmethod
    def with_traceback(cls, code, message, id_):
        data = {"traceback": traceback.format_exc()}
        return cls(code, message, id_, data)



#!/usr/bin/env python3

import asyncio
import contextlib
import logging
import random
import traceback

import asgineer

import not_my_board._jsonrpc as jsonrpc
import not_my_board._util as util

logger = logging.getLogger(__name__)
valid_tokens = ("dummy-token-1", "dummy-token-2")


def serve():
    asgineer.run(asgi_app, "uvicorn", ":2092")


@asgineer.to_asgi
async def asgi_app(request):
    if isinstance(request, asgineer.WebsocketRequest):
        if request.path == "/ws-agent":
            return await _handle_agent(request)
        elif request.path == "/ws-exporter":
            return await _handle_exporter(request)
        await request.close()
        return
    elif isinstance(request, asgineer.HttpRequest):
        if request.path == "/api/v1/places":
            return {"places": [p.desc for p in Place.all()]}
    return 404, {}, "Page not found"


async def _handle_agent(ws):
    await _authorize_ws(ws)
    client_ip = ws.scope["client"][0]
    async with Place.reservation_context(client_ip) as ctx:
        api = AgentAPI(ctx)
        server = jsonrpc.Server(ws.send, ws.receive_iter(), api)
        await server.serve_forever()


class AgentAPI:
    def __init__(self, reservation_context):
        self._reservation_context = reservation_context

    async def reserve(self, candidate_ids):
        place = await Place.reserve(candidate_ids, self._reservation_context)
        return place.desc["id"]

    async def return_reservation(self, place_id):
        await Place.return_by_id(place_id, self._reservation_context)


async def _handle_exporter(ws):
    await _authorize_ws(ws)
    client_ip = ws.scope["client"][0]
    exporter = jsonrpc.Proxy(ws.send, ws.receive_iter())
    async with util.background_task(exporter.io_loop()) as io_loop:
        place = await exporter.get_place()
        with Place.register(place, exporter, client_ip):
            await io_loop


async def _authorize_ws(ws):
    try:
        auth = ws.headers["authorization"]
        scheme, token = auth.split(" ", 1)
        if scheme != "Bearer":
            raise ProtocolError(f"Invalid Authorization Scheme: {scheme}")
        if token not in valid_tokens:
            raise ProtocolError("Invalid token")
    except Exception:
        traceback.print_exc()
        await ws.close()
        return

    await ws.accept()


class Place:
    _all_places = {}
    _next_id = 1
    _available = set()
    _wait_queue = []
    _reservations = {}

    @classmethod
    def all(cls):
        return cls._all_places.values()

    @classmethod
    def get_by_id(cls, id_, default=None):
        return cls._all_places.get(id_, default)

    @classmethod
    def _new_id(cls):
        id_ = cls._next_id
        cls._next_id += 1
        return id_

    @classmethod
    @contextlib.contextmanager
    def register(cls, desc, exporter, client_ip):
        self = cls()
        self._desc = desc
        self._exporter = exporter

        self._id = cls._new_id()
        self._desc["id"] = self._id
        self._desc["host"] = client_ip
        try:
            logger.info("New place registered: %d", self._id)
            cls._all_places[self._id] = self
            cls._available.add(self._id)
            yield self
        finally:
            logger.info("Place disappeared: %d", self._id)
            del cls._all_places[self._id]
            cls._available.discard(self._id)
            for candidates, _, future in cls._wait_queue:
                candidates.discard(self._id)
                if not candidates and not future.done():
                    future.set_exception(Exception("All candidate places are gone"))

    @property
    def desc(self):
        return self._desc

    @classmethod
    @contextlib.asynccontextmanager
    async def reservation_context(cls, client_ip):
        ctx = _ReservationContext(client_ip)
        try:
            cls._reservations[ctx] = set()
            yield ctx
        finally:
            for place in cls._reservations[ctx].copy():
                await cls.return_by_id(place, ctx)
            del cls._reservations[ctx]

    @classmethod
    async def reserve(cls, candidate_ids, ctx):
        existing_candidates = {id_ for id_ in candidate_ids if id_ in cls._all_places}
        if not existing_candidates:
            raise RuntimeError("None of the candidates exist anymore")

        available_candidates = existing_candidates & cls._available
        if available_candidates:
            # TODO do something smart to get the best candidate
            reserved_id = random.choice(list(available_candidates))

            cls._available.remove(reserved_id)
            cls._reservations[ctx].add(reserved_id)
            logger.info("Place reserved: %d", reserved_id)
            place = cls._all_places[reserved_id]
        else:
            logger.debug(
                "No places available, adding request to queue: %s",
                str(existing_candidates),
            )
            future = asyncio.get_running_loop().create_future()
            entry = (existing_candidates, ctx, future)
            cls._wait_queue.append(entry)
            try:
                place = await future
            finally:
                cls._wait_queue.remove(entry)

        # TODO refactor Place class
        # pylint: disable=protected-access
        try:
            await place._exporter.set_allowed_ips([ctx.client_ip])
        except Exception:
            await cls.return_by_id(place._id, ctx)
            raise

        return place

    @classmethod
    async def return_by_id(cls, place_id, ctx):
        cls._reservations[ctx].remove(place_id)
        if place_id in cls._all_places:
            for candidates, new_ctx, future in cls._wait_queue:
                if place_id in candidates and not future.done():
                    cls._reservations[new_ctx].add(place_id)
                    logger.info("Place returned and reserved again: %d", place_id)
                    future.set_result(cls._all_places[place_id])
                    break
            else:
                logger.info("Place returned: %d", place_id)
                cls._available.add(place_id)
                # pylint: disable=protected-access
                await cls._all_places[place_id]._exporter.set_allowed_ips([])
        else:
            logger.info("Place returned, but it doesn't exist: %d", place_id)


class _ReservationContext:
    def __init__(self, client_ip):
        self._client_ip = client_ip

    @property
    def client_ip(self):
        return self._client_ip


class ProtocolError(Exception):
    pass



import urllib.parse

import websockets


def ws_connect(url, auth=None):
    url = urllib.parse.urlsplit(url)

    if url.scheme == "http":
        ws_scheme = "ws"
    elif url.scheme == "https":
        ws_scheme = "wss"
    else:
        ws_scheme = url.scheme

    uri = f"{ws_scheme}://{url.netloc}{url.path}"
    headers = {"Authorization": auth} if auth else {}

    return websockets.connect(uri, extra_headers=headers)



#!/usr/bin/env python3

import asyncio
import contextlib
import logging
import os
import pathlib
import traceback
import urllib.parse

import websockets

import not_my_board._http as http
import not_my_board._jsonrpc as jsonrpc
import not_my_board._models as models
import not_my_board._usbip as usbip
import not_my_board._util as util

logger = logging.getLogger(__name__)


async def agent(server_url):
    async with Agent(server_url) as a:
        await a.serve_forever()


class Agent:
    def __init__(self, server_url):
        self._server_url = server_url
        self._reserved_places = {}
        self._pending = set()

    async def __aenter__(self):
        runtime_dir = pathlib.Path(os.environ["XDG_RUNTIME_DIR"])

        async with contextlib.AsyncExitStack() as stack:
            url = urllib.parse.urlsplit(self._server_url)
            ws_scheme = "ws" if url.scheme == "http" else "wss"
            uri = f"{ws_scheme}://{url.netloc}/ws-agent"
            headers = {"Authorization": "Bearer dummy-token-1"}
            ws = await stack.enter_async_context(
                websockets.connect(uri, extra_headers=headers)
            )

            async def receive_iter():
                try:
                    while True:
                        yield await ws.recv()
                except websockets.ConnectionClosedOK:
                    pass

            self._server_proxy = jsonrpc.Proxy(ws.send, receive_iter())

            stack.push_async_callback(self._cleanup)

            self._unix_server = await stack.enter_async_context(
                util.UnixServer(self._handle_client, runtime_dir / "not-my-board.sock")
            )

            self._stack = stack.pop_all()
            await self._stack.__aenter__()
            return self

    async def __aexit__(self, exc_type, exc, tb):
        await self._stack.__aexit__(exc_type, exc, tb)

    async def _cleanup(self):
        for _, place in self._reserved_places.items():
            if place.is_attached:
                await place.detach()

    # TODO: hide from JSON-RPC interface
    async def serve_forever(self):
        await util.run_concurrently(
            self._unix_server.serve_forever(), self._server_proxy.io_loop()
        )

    async def _handle_client(self, reader, writer):
        async def send(data):
            writer.write(data + b"\n")
            await writer.drain()

        socket_server = jsonrpc.Server(send, reader, self)
        await socket_server.serve_forever()

    async def reserve(self, name, spec_file):
        if name in self._reserved_places:
            raise RuntimeError(f'A place named "{name}" is already reserved')

        if name in self._pending:
            raise RuntimeError(f'A place named "{name}" is currently being reserved')

        self._pending.add(name)
        try:
            spec_content = util.toml_loads(pathlib.Path(spec_file).read_text())
            spec = models.Spec(name=name, **spec_content)

            response = await http.get_json(f"{self._server_url}/api/v1/places")
            places = [models.Place(**p) for p in response["places"]]

            candidates = _filter_places(spec, places)
            candidate_ids = list(candidates)
            if not candidate_ids:
                raise RuntimeError("No matching place found")

            place_id = await self._server_proxy.reserve(candidate_ids)

            assert name not in self._reserved_places
            self._reserved_places[name] = candidates[place_id]
        finally:
            self._pending.remove(name)

    async def return_reservation(self, name, force=False):
        async with self._reserved_place(name) as reserved_place:
            if reserved_place.is_attached:
                if force:
                    await reserved_place.detach()
                else:
                    raise RuntimeError(f'Place "{name}" is still attached')
            await self._server_proxy.return_reservation(reserved_place.id)
            del self._reserved_places[name]

    async def attach(self, name):
        async with self._reserved_place(name) as reserved_place:
            await reserved_place.attach()

    async def detach(self, name):
        async with self._reserved_place(name) as reserved_place:
            await reserved_place.detach()

    @contextlib.asynccontextmanager
    async def _reserved_place(self, name):
        def check_reserved():
            if name not in self._reserved_places:
                raise RuntimeError(f'A place named "{name}" is not reserved')

        check_reserved()
        reserved_place = self._reserved_places[name]
        async with reserved_place.lock:
            # after waiting for the lock the place might have been returned, so
            # check again
            check_reserved()
            yield reserved_place

    async def list(self):
        return [
            {"place": name, "attached": place.is_attached}
            for name, place in self._reserved_places.items()
        ]

    async def status(self):
        usbip.refresh_vhci_status()
        return [
            {"place": name, **status}
            for name, place in self._reserved_places.items()
            for status in place.status
        ]


def _filter_places(spec, places):
    reserved_places = {}

    spec_part_sets = [
        (name, _part_to_set(spec_part)) for name, spec_part in spec.parts.items()
    ]

    for place in places:
        matching = _find_matching(spec_part_sets, place)
        if matching:
            reserved_places[place.id] = ReservedPlace(spec, place, matching)

    return reserved_places


def _find_matching(spec_part_sets, place):
    match_graph = {}
    for name, spec_part_set in spec_part_sets:
        match_graph[name] = []
        for place_part_idx, place_part in enumerate(place.parts):
            place_part_set = _part_to_set(place_part)
            if spec_part_set.issubset(place_part_set):
                match_graph[name].append(place_part_idx)

        if not match_graph[name]:
            return None

    matching = []
    for name, matches in match_graph.items():
        if len(matches) != 1:
            # Complex matching
            matching_dict = util.find_matching(match_graph)
            if len(match_graph) == len(matching_dict):
                return matching_dict.items()
            else:
                return None

        matching.append((name, matches[0]))

    return matching


def _part_to_set(part):
    return (
        {f"compatible:{c}" for c in part.compatible}
        | {f"usb:{k}" for k in part.usb}
        | {f"tcp:{k}" for k in part.tcp}
    )


class ReservedPlace:
    def __init__(self, spec, place, matching):
        self._spec = spec
        self._place = place
        self._tunnels = []
        self._stack = None
        self.lock = asyncio.Lock()
        proxy = str(place.host), place.port

        for name, place_part_idx in matching:
            spec_part = spec.parts[name]
            place_part = place.parts[place_part_idx]

            for usb_name, usb_spec in spec_part.usb.items():
                self._tunnels.append(
                    UsbTunnel(
                        part_name=name,
                        iface_name=usb_name,
                        proxy=proxy,
                        usbid=place_part.usb[usb_name].usbid,
                        vhci_port=usb_spec.vhci_port,
                    )
                )

            for tcp_name, tcp_spec in spec_part.tcp.items():
                self._tunnels.append(
                    TcpTunnel(
                        part_name=name,
                        iface_name=tcp_name,
                        proxy=proxy,
                        remote=(
                            place_part.tcp[tcp_name].host,
                            place_part.tcp[tcp_name].port,
                        ),
                        local_port=tcp_spec.local_port,
                    )
                )

    async def attach(self):
        if self._stack is not None:
            raise RuntimeError('Place "{self._spec.name}" is already attached')

        async with contextlib.AsyncExitStack() as stack:
            for tunnel in self._tunnels:
                await stack.enter_async_context(tunnel)
            self._stack = stack.pop_all()

    async def detach(self):
        if self._stack is None:
            raise RuntimeError('Place "{self._spec.name}" is not attached')

        try:
            await self._stack.aclose()
        finally:
            self._stack = None

    @property
    def id(self):
        return self._place.id

    @property
    def is_attached(self):
        return self._stack is not None

    @property
    def status(self):
        return [
            {
                "part": t.part_name,
                "interface": t.iface_name,
                "type": t.type_name,
                "attached": t.attached,
            }
            for t in self._tunnels
        ]


class UsbTunnel:
    _target = "usb.not-my-board.localhost", 3240
    _ready_timeout = 5

    def __init__(self, part_name, iface_name, proxy, usbid, vhci_port):
        self._part_name = part_name
        self._iface_name = iface_name
        self._name = f"{part_name}.{iface_name}"
        self._proxy = proxy
        self._usbid = usbid
        self._vhci_port = vhci_port

    async def __aenter__(self):
        async with contextlib.AsyncExitStack() as stack:
            ready_event = asyncio.Event()
            await stack.enter_async_context(
                util.background_task(self._tunnel_task(ready_event))
            )
            logger.debug("%s: Attaching USB device", self._name)

            try:
                await asyncio.wait_for(ready_event.wait(), self._ready_timeout)
            except TimeoutError:
                logger.warning("%s: Attaching USB device timed out", self._name)

            self._stack = stack.pop_all()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self._stack.__aexit__(exc_type, exc, tb)

    async def _tunnel_task(self, ready_event):
        retry_timeout = 1
        try:
            while True:
                try:
                    await self._attach()
                    ready_event.set()
                    retry_timeout = 1
                except Exception:
                    traceback.print_exc()
                    await asyncio.sleep(retry_timeout)
                    retry_timeout = min(2 * retry_timeout, 30)
        finally:
            usbip.detach(self._vhci_port)
            logger.debug("%s: USB device detached", self._name)

    async def _attach(self):
        tunnel = http.open_tunnel(*self._proxy, *self._target)
        async with tunnel as (reader, writer, trailing_data):
            if trailing_data:
                raise ProtocolError("USB/IP implementation cannot handle trailing data")
            await usbip.attach(reader, writer, self._usbid, self._vhci_port)
        logger.debug("%s: USB device attached", self._name)

    @property
    def part_name(self):
        return self._part_name

    @property
    def iface_name(self):
        return self._iface_name

    @property
    def type_name(self):
        return "USB"

    @property
    def attached(self):
        return usbip.is_attached(self._vhci_port)


class TcpTunnel:
    def __init__(self, part_name, iface_name, proxy, remote, local_port):
        self._part_name = part_name
        self._iface_name = iface_name
        self._name = f"{part_name}.{iface_name}"
        self._proxy = proxy
        self._remote = remote
        self._local_port = local_port
        self._is_attached = False

    async def __aenter__(self):
        async with contextlib.AsyncExitStack() as stack:
            ready_event = asyncio.Event()
            await stack.enter_async_context(
                util.background_task(self._tunnel_task(ready_event))
            )
            await ready_event.wait()
            self._is_attached = True
            self._stack = stack.pop_all()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self._stack.__aexit__(exc_type, exc, tb)
        self._is_attached = False

    async def _tunnel_task(self, ready_event):
        localhost = "127.0.0.1"
        async with util.Server(
            self._handle_client, localhost, self._local_port
        ) as server:
            ready_event.set()
            await server.serve_forever()

    async def _handle_client(self, client_r, client_w):
        logger.debug("%s: Opening tunnel", self._name)
        async with http.open_tunnel(*self._proxy, *self._remote) as (
            remote_r,
            remote_w,
            trailing_data,
        ):
            logger.debug("%s: Tunnel created, relaying data", self._name)
            client_w.write(trailing_data)
            await client_w.drain()
            await util.relay_streams(client_r, client_w, remote_r, remote_w)

    @property
    def part_name(self):
        return self._part_name

    @property
    def iface_name(self):
        return self._iface_name

    @property
    def type_name(self):
        return "TCP"

    @property
    def attached(self):
        return self._is_attached


class ProtocolError(Exception):
    pass



#!/usr/bin/env python3

import asyncio
import contextlib
import dataclasses
import logging
import os
import pathlib
import socket
import struct
import sys

import not_my_board._util as util

if sys.version_info < (3, 9):
    from typing_extensions import Annotated
else:
    from typing import Annotated


logger = logging.getLogger(__name__)
_vhci_status_attached = {}


class UsbIpServer:
    def __init__(self, devices):
        self._devices = {d.busid: d for d in devices}

    async def __aenter__(self):
        async with contextlib.AsyncExitStack() as stack:
            for _, device in self._devices.items():
                await stack.enter_async_context(
                    util.background_task(_refresh_task(device))
                )

            self._stack = stack.pop_all()
            await self._stack.__aenter__()
            return self

    async def __aexit__(self, exc_type, exc, tb):
        await self._stack.__aexit__(exc_type, exc, tb)

    async def handle_client(self, reader, writer):
        sock = writer.transport.get_extra_info("socket")
        _enable_keep_alive(sock)

        request = await ImportRequest.from_reader(reader)
        logger.debug("Received: %s", request)
        if request.busid not in self._devices:
            raise ProtocolError(f"Unexpected Bus ID: {request.busid}")

        device = self._devices[request.busid]

        # allow client to trigger a refresh
        device.refresh()

        async with device:
            writer.transport.pause_reading()
            fd = sock.fileno()
            device.export(fd)

            reply = ImportReply.from_device(device)
            logger.debug("Sending: %s", reply)
            writer.write(bytes(reply))
            await writer.drain()

            writer.close()
            await writer.wait_closed()
            await device.available()


class _SysfsFileInt:
    def __set_name__(self, owner, name):
        self._name = name

    def __init__(self, base=10, default=None):
        self._base = base
        self._default = default

    def __get__(self, instance, owner=None):
        try:
            path = instance._sysfs_path / self._name
            return int(path.read_text(), base=self._base)
        except ValueError:
            if self._default is not None:
                return self._default
            else:
                raise


class _SysfsFileHex(_SysfsFileInt):
    def __init__(self, default=None):
        super().__init__(16, default)


class UsbIpDevice:
    def __init__(self, busid):
        self._busid = busid
        self._sysfs_path = pathlib.Path("/sys/bus/usb/devices/") / busid
        self._lock = asyncio.Lock()
        self._refresh_event = asyncio.Event()
        self._is_exported = False

    def refresh(self):
        self._refresh_event.set()

    async def __aenter__(self):
        async with contextlib.AsyncExitStack() as stack:
            await stack.enter_async_context(self._lock)
            await self.available()

            self._stack = stack.pop_all()
            await self._stack.__aenter__()
            return self

    async def __aexit__(self, exc_type, exc, tb):
        if self._is_exported:
            try:
                (self._sysfs_path / "usbip_sockfd").write_text("-1\n")
            except (OSError, FileNotFoundError):
                # client might have disconnected or device disappeared
                pass
            except Exception as e:
                logger.warning("Error while stopping export: %s", e)

        await self._stack.__aexit__(exc_type, exc, tb)

    async def available(self):
        while not self._is_available():
            self._refresh_event.clear()
            await self._refresh_event.wait()

    def export(self, fd):
        (self._sysfs_path / "usbip_sockfd").write_text(f"{fd}\n")
        self._is_exported = True

    def _is_available(self):
        try:
            status_available = 1
            if self.usbip_status == status_available:
                logger.debug("Device %s is available", self.busid)
                return True
        except FileNotFoundError:
            # device might have disappeared
            pass
        except Exception as e:
            logger.warning("Error while checking device status: %s", e)

        logger.debug("Device %s is not available, yet", self.busid)
        return False

    @property
    def busid(self):
        return self._busid.encode("utf-8")

    @property
    def path(self):
        return self._sysfs_path.as_posix().encode("utf-8")

    @property
    def speed(self):
        string_to_code = {
            "1.5": 1,
            "12": 2,
            "480": 3,
            "53.3-480": 4,
            "5000": 5,
        }
        string = (self._sysfs_path / "speed").read_text()[:-1]  # strip newline
        return string_to_code.get(string, 0)

    usbip_status = _SysfsFileInt()
    busnum = _SysfsFileInt()
    devnum = _SysfsFileInt()
    idVendor = _SysfsFileHex()
    idProduct = _SysfsFileHex()
    bcdDevice = _SysfsFileHex()
    bDeviceClass = _SysfsFileHex()
    bDeviceSubClass = _SysfsFileHex()
    bDeviceProtocol = _SysfsFileHex()
    bConfigurationValue = _SysfsFileHex(default=0)
    bNumConfigurations = _SysfsFileHex()
    bNumInterfaces = _SysfsFileHex(default=0)


async def attach(reader, writer, busid, port):
    sock = writer.transport.get_extra_info("socket")
    # Client waits 2 seconds longer before sending keep alive probes, otherwise
    # both sides start sending at the same time.
    _enable_keep_alive(sock, extra_idle_sec=2)

    request = ImportRequest(busid.encode())
    logger.debug("Sending: %s", request)
    writer.write(bytes(request))
    await writer.drain()

    reply = await ImportReply.from_reader(reader)
    logger.debug("Received: %s", reply)

    fd = os.dup(sock.fileno())

    try:
        writer.close()
        await writer.wait_closed()

        attach_path = pathlib.Path("/sys/devices/platform/vhci_hcd.0/attach")
        devid = (reply.busnum << 16) | reply.devnum
        attach_path.write_text(f"{port} {fd} {devid} {reply.speed}\n")
    finally:
        os.close(fd)


def detach(port):
    detach_path = pathlib.Path("/sys/devices/platform/vhci_hcd.0/detach")
    try:
        detach_path.write_text(f"{port}")
    except OSError:
        # not attached anymore
        pass


def refresh_vhci_status():
    status_path = pathlib.Path("/sys/devices/platform/vhci_hcd.0/status")
    status_attached = 6  # VDEV_ST_USED
    with status_path.open() as f:
        # skip header:
        # hub port sta spd dev      sockfd local_busid
        f.readline()

        for line in f:
            entries = line.split()
            port = int(entries[1])
            status = int(entries[2])
            _vhci_status_attached[port] = status == status_attached
    # TODO parse other status files if Kernel is compiled with more vhci
    # ports


def is_attached(port):
    return _vhci_status_attached[port]


def _enable_keep_alive(sock, extra_idle_sec=0):
    # enable TCP keep alive
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)

    # Drop connections faster than the 2 hour default: Send first probe after 5
    # (+extra_idle_sec) seconds, then every 5 seconds. After 3 unanswered
    # probes the connection is closed.
    sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, 5 + extra_idle_sec)
    sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, 5)
    sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, 3)


class StructType:
    def __init__(self, format_str):
        self.format_str = format_str


class StructStr(StructType):
    pass


UInt8 = Annotated[int, StructType("B")]
UInt16 = Annotated[int, StructType("H")]
UInt32 = Annotated[int, StructType("I")]


class Char(bytes):
    def __class_getitem__(cls, key):
        return Annotated[bytes, StructStr(f"{key}s")]


# pylint: disable=W0212,E1101
# - accessing protected member of cls, e.g. cls._struct
# - 'serializable' has no '_struct' member
def serializable(cls):
    cls = dataclasses.dataclass(cls)

    format_str = "!"
    to_strip = set()

    for field in dataclasses.fields(cls):
        for metadata in field.type.__metadata__:
            if isinstance(metadata, StructType):
                format_str += metadata.format_str
                if isinstance(metadata, StructStr):
                    to_strip.add(field.name)
                break
        else:
            raise ProtocolError(f"Field {field!r} not annotated with StructType")

    cls._struct = struct.Struct(format_str)
    cls._to_strip = to_strip

    def __bytes__(self):
        return self._struct.pack(
            *(getattr(self, f.name) for f in dataclasses.fields(self))
        )

    @classmethod
    async def from_reader(cls, reader):
        data = await reader.readexactly(cls._struct.size)
        values = cls._struct.unpack(data)
        init_values = []
        for field, value in zip(dataclasses.fields(cls), values):
            if field.init:
                if field.name in cls._to_strip:
                    value = value.rstrip(b"\0")
                init_values.append(value)
            else:
                if value != field.default:
                    raise ProtocolError(
                        f"Expected {field.name}={field.default}, got={value}"
                    )

        # pylint: disable=E1120
        # No value for argument 'cls': false positive
        return cls(*init_values)

    cls.__bytes__ = __bytes__
    cls.from_reader = from_reader

    return cls


def no_init(default):
    return dataclasses.field(default=default, init=False)


@serializable
class Header:
    version: UInt16 = no_init(0x0111)
    code: UInt16
    status: UInt32 = no_init(0)


@serializable
class ImportRequest(Header):
    code: UInt16 = no_init(0x8003)
    busid: Char[32]


@serializable
class ImportReply(Header):
    code: UInt16 = no_init(0x0003)
    path: Char[256]
    busid: Char[32]
    busnum: UInt32
    devnum: UInt32
    speed: UInt32
    idVendor: UInt16
    idProduct: UInt16
    bcdDevice: UInt16
    bDeviceClass: UInt8
    bDeviceSubClass: UInt8
    bDeviceProtocol: UInt8
    bConfigurationValue: UInt8
    bNumConfigurations: UInt8
    bNumInterfaces: UInt8

    @classmethod
    def from_device(cls, device):
        return cls(
            *(
                getattr(device, field.name)
                for field in dataclasses.fields(cls)
                if field.init
            )
        )


class ProtocolError(Exception):
    pass


async def _refresh_task(device):
    pipe_path = pathlib.Path("/run/usbip-refresh-" + device.busid.decode())

    tmp_path = pipe_path.with_name(pipe_path.name + ".new")
    os.mkfifo(tmp_path)
    tmp_path.replace(pipe_path)

    async with _open_read_pipe(pipe_path, "r+b", buffering=0) as pipe:
        while True:
            await pipe.read(4096)
            device.refresh()


@contextlib.asynccontextmanager
async def _open_read_pipe(*args, **kwargs):
    with open(*args, **kwargs) as pipe:
        loop = asyncio.get_running_loop()
        reader = asyncio.StreamReader(loop=loop)
        protocol = asyncio.StreamReaderProtocol(reader, loop=loop)
        await loop.connect_read_pipe(lambda: protocol, pipe)
        yield reader


# pylint: disable=import-outside-toplevel
async def _main():
    import argparse

    logging.basicConfig(
        format="%(asctime)s %(levelname)s: %(name)s: %(message)s", level=logging.DEBUG
    )

    parser = argparse.ArgumentParser(description="Import and export USB ports")
    subparsers = parser.add_subparsers(dest="command", required=True, metavar="command")

    subparser = subparsers.add_parser("export", help="export a USB device")
    subparser.add_argument(
        "busid", help='busid of the device to export, e.g. "1-5.1.4"'
    )
    subparser.add_argument("-p", "--port", default=3240, help="port to listen on")

    subparser = subparsers.add_parser("import", help="import a USB device")
    subparser.add_argument("host", help="host to connect to")
    subparser.add_argument(
        "busid", help='busid of the device to import, e.g. "1-5.1.4"'
    )
    subparser.add_argument("vhci_port", help='vhci port to attach device to, e.g. "0"')
    subparser.add_argument("-p", "--port", default=3240, help="port to connect to")

    args = parser.parse_args()

    if args.command == "import":
        while True:
            logger.info("Connecting")
            reader, writer = await asyncio.open_connection(
                args.host, args.port, family=socket.AF_INET
            )
            await attach(reader, writer, args.busid, args.vhci_port)
    else:
        device = UsbIpDevice(args.busid)
        async with UsbIpServer([device]) as usbip_server:
            server = util.Server(
                usbip_server.handle_client, port=args.port, family=socket.AF_INET
            )
            async with server:
                logger.info("listening")
                await server.serve_forever()


if __name__ == "__main__":
    util.run(_main())



#!/usr/bin/python3

import contextlib
import json
import urllib.parse

import h11

import not_my_board._util as util


class ProtocolError(Exception):
    pass


async def get_json(url):
    url = urllib.parse.urlsplit(url)
    headers = [
        ("Host", url.netloc),
        ("User-Agent", h11.PRODUCT_ID),
        ("Accept", "application/json"),
        ("Connection", "close"),
    ]

    conn = h11.Connection(our_role=h11.CLIENT)
    to_send = conn.send(
        h11.Request(method="GET", target=url.path or "/", headers=headers)
    )

    if url.scheme == "https":
        default_port = 443
        ssl = True
    elif url.scheme == "http":
        default_port = 80
        ssl = False
    else:
        raise ValueError("Unknown scheme '{url.scheme}'")

    port = url.port or default_port

    async with util.connect(url.hostname, port, ssl=ssl) as (reader, writer):
        writer.write(to_send)
        writer.write(conn.send(h11.EndOfMessage()))
        await writer.drain()

        async def receive_all():
            while True:
                event = conn.next_event()
                if event is h11.NEED_DATA:
                    conn.receive_data(await reader.read(4096))
                elif isinstance(event, h11.Response):
                    if event.status_code != 200:
                        raise ProtocolError(
                            f"Expected status code 200, got {event.status_code}"
                        )
                elif isinstance(event, h11.Data):
                    yield event.data
                elif isinstance(event, (h11.EndOfMessage, h11.PAUSED)):
                    break

        content = b"".join([data async for data in receive_all()])

    return json.loads(content)


@contextlib.asynccontextmanager
async def open_tunnel(proxy_host, proxy_port, target_host, target_port):
    headers = [
        ("Host", f"{target_host}:{target_port}"),
        ("User-Agent", h11.PRODUCT_ID),
    ]

    conn = h11.Connection(our_role=h11.CLIENT)
    to_send = conn.send(
        h11.Request(
            method="CONNECT", target=f"{target_host}:{target_port}", headers=headers
        )
    )

    async with util.connect(proxy_host, proxy_port) as (reader, writer):
        writer.write(to_send)
        writer.write(conn.send(h11.EndOfMessage()))
        await writer.drain()

        while True:
            event = conn.next_event()
            if event is h11.NEED_DATA:
                conn.receive_data(await reader.read(4096))
            elif isinstance(event, h11.Response):
                response = event
                if response.status_code != 200:
                    raise ProtocolError(
                        f"Expected status code 200, got {event.status_code}"
                    )

                yield reader, writer, conn.trailing_data[0]
                break
            else:
                raise ProtocolError(f"Unexpected event: {event}")



import ipaddress
from typing import Dict, List

import pydantic

UsbId = pydantic.constr(regex=r"^[1-9][0-9]*-[1-9][0-9]*(\.[1-9][0-9]*)*$")


class UsbSpec(pydantic.BaseModel):
    vhci_port: pydantic.NonNegativeInt


class TcpSpec(pydantic.BaseModel):
    local_port: pydantic.PositiveInt


class SpecPart(pydantic.BaseModel):
    compatible: List[str]
    usb: Dict[str, UsbSpec] = {}
    tcp: Dict[str, TcpSpec] = {}


class Spec(pydantic.BaseModel):
    name: str
    parts: Dict[str, SpecPart]


class UsbDesc(pydantic.BaseModel):
    usbid: UsbId


class TcpDesc(pydantic.BaseModel):
    host: str
    port: pydantic.PositiveInt


class ExportedPart(pydantic.BaseModel):
    compatible: List[str]
    usb: Dict[str, UsbDesc] = {}
    tcp: Dict[str, TcpDesc] = {}


class ExportDesc(pydantic.BaseModel):
    port: pydantic.PositiveInt
    parts: List[ExportedPart]


class Place(ExportDesc):
    id: pydantic.PositiveInt
    host: ipaddress.IPv4Address



#!/usr/bin/env python3

import asyncio
import contextlib
import datetime
import email.utils
import ipaddress
import logging
import socket

import h11
import websockets

import not_my_board._jsonrpc as jsonrpc
import not_my_board._models as models
import not_my_board._usbip as usbip
import not_my_board._util as util

logger = logging.getLogger(__name__)


async def export(server_url, place):
    async with Exporter(server_url, place) as exporter:
        await exporter.serve_forever()


class Exporter:
    def __init__(self, server_url, export_desc_path):
        self._server_url = server_url
        self._ip_to_tasks_map = {}
        export_desc_content = export_desc_path.read_text()
        self._place = models.ExportDesc(**util.toml_loads(export_desc_content))

        tcp_targets = {
            f"{tcp.host}:{tcp.port}".encode()
            for part in self._place.parts
            for _, tcp in part.tcp.items()
        }
        self._usb_target = {b"usb.not-my-board.localhost:3240"}
        self._allowed_proxy_targets = tcp_targets | self._usb_target

        self._usbip_devices = [
            usbip.UsbIpDevice(usb.usbid)
            for part in self._place.parts
            for _, usb in part.usb.items()
        ]

    async def __aenter__(self):
        async with contextlib.AsyncExitStack() as stack:
            self._usbip_server = await stack.enter_async_context(
                usbip.UsbIpServer(self._usbip_devices)
            )

            self._http_server = await stack.enter_async_context(
                util.Server(
                    self._handle_client, port=self._place.port, family=socket.AF_INET
                )
            )

            url = f"{self._server_url}/ws-exporter"
            auth = "Bearer dummy-token-1"
            self._ws = await stack.enter_async_context(util.ws_connect(url, auth))
            self._ws_server = jsonrpc.Server(self._ws.send, self._receive_iter(), self)

            self._stack = stack.pop_all()
            await self._stack.__aenter__()
            return self

    async def __aexit__(self, exc_type, exc, tb):
        await self._stack.__aexit__(exc_type, exc, tb)

    # TODO: hide from JSON-RPC interface
    async def serve_forever(self):
        await util.run_concurrently(
            self._http_server.serve_forever(), self._ws_server.serve_forever()
        )

    async def _receive_iter(self):
        try:
            while True:
                yield await self._ws.recv()
        except websockets.ConnectionClosedOK:
            pass

    async def get_place(self):
        return self._place.dict()

    async def set_allowed_ips(self, ips):
        new_ips = set(map(ipaddress.ip_address, ips))
        old_ips = set(self._ip_to_tasks_map)

        removed_ips = old_ips - new_ips
        added_ips = new_ips - old_ips

        if removed_ips:
            logger.info("Allowed IPs removed: %s", ", ".join(map(str, removed_ips)))

        if added_ips:
            logger.info("Allowed IPs added: %s", ", ".join(map(str, added_ips)))

        to_close = [
            task for ip in removed_ips for task in self._ip_to_tasks_map.pop(ip)
        ]

        for ip in added_ips:
            self._ip_to_tasks_map[ip] = set()

        await util.cancel_tasks(to_close)

    async def _handle_client(self, reader, writer):
        con = HttpProxyConnection(reader, writer, self._allowed_proxy_targets)
        host, _ = writer.transport.get_extra_info("peername")
        client_ip = ipaddress.ip_address(host)

        if client_ip in self._ip_to_tasks_map:
            task = asyncio.current_task()
            tasks = self._ip_to_tasks_map[client_ip]
            tasks.add(task)
            task.add_done_callback(tasks.discard)

            target, trailing_data = await con.receive_target()
            logger.info("Proxy CONNECT target: %s", target)

            await self._tunnel(reader, writer, target, trailing_data)
        else:
            await con.deny_request()

    async def _tunnel(self, client_r, client_w, target, trailing_data):
        if target == b"usb.not-my-board.localhost:3240":
            if trailing_data:
                raise ProtocolError("USB/IP implementation cannot handle trailing data")
            await self._usbip_server.handle_client(client_r, client_w)
        else:
            host, port = target.split(b":", 1)
            port = int(port)

            async with util.connect(host, port) as (remote_r, remote_w):
                remote_w.write(trailing_data)
                await remote_w.drain()
                await util.relay_streams(client_r, client_w, remote_r, remote_w)


def format_date_time(dt=None):
    """Generate a RFC 7231 / RFC 9110 IMF-fixdate string"""
    if dt is None:
        dt = datetime.datetime.now(datetime.timezone.utc)
    return email.utils.format_datetime(dt, usegmt=True)


class HttpProxyConnection:
    _msg_wrong_method = (
        b"This is a not-my-board export server. "
        b"You probably want to use not-my-board, instead of connecting directly.\n"
    )
    _msg_wrong_ip = (
        b"This is a not-my-board export server. Your IP address is not allowed.\n"
    )
    _msg_wrong_target = (
        b"This is a not-my-board export server. "
        b"The requested target is not allowed.\n"
    )

    def __init__(self, reader, writer, allowed_targets):
        self._conn = h11.Connection(h11.SERVER)
        self._reader = reader
        self._writer = writer
        self._allowed_targets = allowed_targets

    async def deny_request(self):
        await self._send_response(403, self._msg_wrong_ip)

    async def receive_target(self):
        try:
            while True:
                event = self._conn.next_event()
                if event is h11.NEED_DATA:
                    data = await self._reader.read(8 * 1024)
                    self._conn.receive_data(data)
                elif isinstance(event, h11.Request):
                    request = event
                    if request.method == b"CONNECT":
                        if request.target in self._allowed_targets:
                            await self._send_response(200)
                            return request.target, self._conn.trailing_data[0]
                        else:
                            await self._send_response(403, self._msg_wrong_target)
                            raise ProtocolError(
                                f"Forbidden target requested: {request.target}"
                            )
                    else:
                        headers = [("Allow", "CONNECT")]
                        await self._send_response(405, self._msg_wrong_method, headers)
                        raise ProtocolError(f"Unexpected Method: {request.method}")
                else:
                    raise ProtocolError(f"Unexpected Event: {event}")
        except Exception as e:
            if self._conn.our_state in {h11.IDLE, h11.SEND_RESPONSE}:
                if isinstance(e, h11.RemoteProtocolError):
                    status_code = e.error_status_hint
                else:
                    status_code = 500
                await self._send_response(status_code, body=str(e).encode())
            raise

    async def _send_response(self, status, body=None, headers=None, content_type=None):
        all_headers = [
            ("Date", format_date_time()),
            ("Server", h11.PRODUCT_ID),
            ("Connection", "close"),
        ]

        if body is not None:
            if not content_type:
                content_type = "text/plain"
            all_headers += [
                ("Content-Type", f"{content_type}; charset=utf-8"),
                ("Content-Length", str(len(body))),
            ]

        if headers:
            all_headers += headers

        res = h11.Response(status_code=status, headers=all_headers)
        end = h11.EndOfMessage()

        if body is not None:
            await self._send([res, h11.Data(data=body), end])
        else:
            if status != 200:
                await self._send([res, end])
            else:
                await self._send([res])

    async def _send(self, events):
        data = b"".join([self._conn.send(event) for event in events])
        self._writer.write(data)
        await self._writer.drain()


class ProtocolError(Exception):
    pass
